/*

			Bullets  --  The Screen Saver
			
				 by John Walker
		
*/ 

#include "bullets.h"

//	Our formidable arsenal

struct weapon weapons[ARSENAL + 1] = {
	{ IDR_PISTOL, 15, 3, 10, 8 },
	{ IDR_BUDDA, 19, 12, 30, 15 },
	{ IDR_RIFLE, 31, 4, 15, 6 },
	{ IDR_CARBINE, 28, 4,  20, 8 },
	{ IDR_SHOTGUN, 40, 2, 3, 4 },
	{ IDR_SHOTECHO, 43, 2, 3, 3 },
	{ IDR_CANNON, 74, 1, 5, 4 },
	{ IDR_HOWITZER, 80, 1, 5, 4 },

	{ IDR_HANGFIRE, 0, 1, 0, 0 }			// This line must be the last
};

//  INITSEED  --  Generate initial random seed.

static void initseed(void)
{
	srandom(GetTickCount());
}

//  GUNSHOT  --  Make the gunshot sound of the chosen weapon.

static void gunshot(int weap)
{
	HANDLE hResInfo, hRes;
	LPSTR lpRes;

	if (soundGunfire && (weapons[weap].sound > 0)) {
		hResInfo = FindResource(hInst, MAKEINTRESOURCE(weapons[weap].sound), "WAVE");
		if (hResInfo != NULL) {
			hRes = LoadResource(hInst, hResInfo);
			if (hRes != NULL) {
				lpRes = LockResource(hRes);
				if (lpRes != NULL) {
					sndPlaySound(lpRes, SND_MEMORY | SND_ASYNC | SND_NODEFAULT);
					UnlockResource(hRes);
				}
			}
			FreeResource(hRes);
		}
	}
}

//  SAVERMAKE	--  Generate next screen saver event.

BOOL saverMake(HWND hWnd, HDC hDC, int gn, int burstN, RECT rc)
{
	int gi;
	RECT r;
	static int cx, cy, lgi;
	BOOL changed = TRUE;
	int scrX, scrY;
	
	scrX = rc.right - rc.left;
	scrY = rc.bottom - rc.top;

	if (burstN > 0) {
		gi = lgi;

		/* Make the next shots in a burst cluster around the last
		   hole.  "I like the pump.  I like the spray." */

		cx += (random() % (gi * 6)) - (gi * 3);
		if (cx < -gi) {
			cx = 0;
		} else if (cx > scrX) {
			cx = scrX - gi;
		}
		/*	Might want to introduce a weapon-dependent upward
			bias here due to muzzle climb.  */
		cy += (random() % (gi * 6)) - (gi * 3);
		if (cy < -gi) {
			cy = 0;
		} else if (cy > scrY) {
			cy = scrY - gi;
		}
	} else {   
		lgi = gi = max(10, (weapons[gn].holesize * max(scrX, scrY)) / 1024);		
		cx = (random() % (scrX + gi)) - gi;
		cy = (random() % (scrY + gi)) - gi;
	}
	if (weapons[gn].holesize > 0) {
		HBITMAP bo = NULL, bn = NULL;
		LPBYTE obits = NULL, nbits = NULL;
		HBRUSH obr;
		HPEN open;

		int ex = cx, ey = cy, ox = 0, oy = 0, gix = gi, giy = gi;
		r.left = r.top = 0;
		r.right = r.bottom = gi;		

			
		if (cx < 0) {
			ox = -cx;
			cx = 0;
			gix -= ox;
		}
		if (cy < 0) {
			oy = -cy;
			cy = 0;
			giy -= oy; 
		}
 

		bo = CreateCompatibleBitmap(hDC, gix, giy);
		if (bo != NULL) {
			HDC hCDC = CreateCompatibleDC(hDC);
			if (hCDC != NULL) {
				HBITMAP obp = SelectObject(hCDC, bo);

				BitBlt(hCDC, 0, 0, gix, giy,
					   hDC, rc.left + cx, rc.top + cy, SRCCOPY);

				SelectObject(hCDC, obp);
				DeleteDC(hCDC);
			}
		}
		
		obr = SelectObject(hDC, GetStockObject(BLACK_BRUSH));
		open = SelectObject(hDC, GetStockObject(BLACK_PEN));
		Ellipse(hDC, rc.left + ex, rc.top + ey, rc.left + ex + gi, rc.top + ey + gi);

		bn = CreateCompatibleBitmap(hDC, gix, giy);
		if (bn != NULL) {
			HDC hCDC = CreateCompatibleDC(hDC);
			if (hCDC != NULL) {
				HBITMAP obp = SelectObject(hCDC, bn);

				BitBlt(hCDC, 0, 0, gix, giy,
					   hDC, rc.left + cx, rc.top + cy, SRCCOPY);

				SelectObject(hCDC, obp);
				DeleteDC(hCDC);
			}
		}
		
		SelectObject(hDC, open);
		SelectObject(hDC, obr); 

		/*	This rather odd piece of code serves to detect whether
			the shot we just fired actually changed anything or just
			went through a hole we've already blown in the screen.
			To do this efficiently, we make before and after copies
			of the region of the hole, using device dependent bitmap
			format, then binary compare them for equality.  If they
			match, the shot didn't connect and we suppress the sound
			for it.  This makes the screen saver eventually shut up as
			the entire screen is cleared. */

		if (bo != NULL && bn != NULL) {
			int i, j;
			BITMAP omap, nmap;

			changed = FALSE;
			GetObject(bo, sizeof omap, &omap);
			GetObject(bn, sizeof nmap, &nmap);
			obits = (LPBYTE) GlobalAlloc(GPTR, omap.bmWidthBytes * omap.bmHeight);
			nbits = (LPBYTE) GlobalAlloc(GPTR, nmap.bmWidthBytes * nmap.bmHeight);
			if (obits != NULL && nbits != NULL) {
				GetBitmapBits(bo, omap.bmWidthBytes * omap.bmHeight, obits);
				GetBitmapBits(bn, nmap.bmWidthBytes * nmap.bmHeight, nbits);
				for (i = 0; i < omap.bmHeight; i++) {
					LPBYTE ob = obits + omap.bmWidthBytes * i,
						   nb = nbits + nmap.bmWidthBytes * i;
					for (j = 0; j < (omap.bmWidth * omap.bmBitsPixel) / 8;
						 j++) {
						if (*ob++ != *nb++) {
							changed = TRUE;
							i = omap.bmHeight;
							break;
						} 
					}
				}
			}
		}

		if (bo != NULL) {
			DeleteObject(bo);
		}
		if (bn != NULL) {
			DeleteObject(bn);
		}
		if (obits != NULL) {
			GlobalFree(obits);
		}
		if (nbits != NULL) {
			GlobalFree(nbits);
		}
	}
	if (changed) {
		gunshot(gn);
	}
#ifdef _DEBUG
	else gunshot(ARSENAL);			// To debug no-change logic
#endif
	return TRUE;
}

//  SAVERINIT  --  Initialise generation.

BOOL saverInit(HWND hWnd)
{
	initseed();
#ifdef _DEBUG
	/* Setting the window non-topmost allows the debugger and
	   BoundsChecker to pop up if something untoward happens
	   while we're running.  Note that this causes a double screen
	   clear on certain displays--not to worry; all will be well
	   (or at least as well as anything can be within the diseased
	   corpse of Windows) in a production build. */
	SetWindowPos(hWnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
#endif
	return TRUE;
}

//  SAVERTERM  --  Terminate and release all resources.

void saverTerm(void)
{
	/*	We don't need no system handles,
		We don't need no thread control,
		Dark sarcasm in the dungeon,
		Windows!  Leave our code alone!  */
}
